#!/usr/bin/env python

import sys

import json

from charmhelpers.core.hookenv import (
    Hooks,
    UnregisteredHookError,
    config,
    local_unit,
    log,
    relation_get,
    relation_ids,
    relation_set
)

from charmhelpers.core.host import (
    restart_on_change,
    service_restart
)

from charmhelpers.fetch import (
    apt_install,
    apt_upgrade,
    configure_sources
)

import contrail_control_utils as utils
from contrail_control_utils import (
    control_network_ip,
    dpkg_version,
    fix_nodemgr,
    fix_permissions,
    provision_control,
    remove_ssl_ca_certificate,
    units,
    unprovision_control,
    write_control_config,
    write_nodemgr_config,
    write_ssl_ca_certificate,
    write_vnc_api_config
)

PACKAGES = [ "contrail-control", "contrail-utils", "contrail-nodemgr" ]

hooks = Hooks()
config = config()

def add_control():
    # check relation dependencies
    if not config_get("control-provisioned") \
       and config_get("contrail-api-ready") \
       and config_get("contrail-discovery-ready") \
       and config_get("contrail-ifmap-ready") \
       and config_get("identity-admin-ready"):
        provision_control()
        config["control-provisioned"] = True

@hooks.hook("config-changed")
def config_changed():
    write_config()
    configure_ssl()
    if config.get("control-provisioned"):
        configure_control_network()

    settings = { "private-address": control_network_ip() }
    for rid in relation_ids("control-node"):
        relation_set(relation_id=rid, relation_settings=settings)

def config_get(key):
    try:
        return config[key]
    except KeyError:
        return None

def configure_control_network():
    if control_network_ip() != config["provisioned-host-ip"]:
        unprovision_control()
        provision_control()

def configure_ssl():
    cert = config.get("ssl-ca")
    if cert:
        write_ssl_ca_certificate(cert)
    else:
        remove_ssl_ca_certificate()

@hooks.hook("contrail-api-relation-changed")
def contrail_api_changed():
    if not relation_get("port"):
        log("Relation not ready")
        return
    write_vnc_api_config()
    config["contrail-api-ready"] = True
    add_control()

@hooks.hook("contrail-api-relation-departed")
@hooks.hook("contrail-api-relation-broken")
def contrail_api_departed():
    if not units("contrail-api"):
        remove_control()
        config["contrail-api-ready"] = False
    write_vnc_api_config()

@hooks.hook("contrail-discovery-relation-changed")
def contrail_discovery_changed():
    if not relation_get("port"):
        log("Relation not ready")
        return
    contrail_discovery_relation()
    config["contrail-discovery-ready"] = True
    add_control()

@hooks.hook("contrail-discovery-relation-departed")
@hooks.hook("contrail-discovery-relation-broken")
def contrail_discovery_departed():
    if not units("contrail-discovery"):
        remove_control()
        config["contrail-discovery-ready"] = False
    contrail_discovery_relation()

@restart_on_change({"/etc/contrail/contrail-control.conf": ["contrail-control"],
                    "/etc/contrail/control-node.conf": ["contrail-control"],
                    "/etc/contrail/contrail-control-nodemgr.conf": ["contrail-control-nodemgr"]})
def contrail_discovery_relation():
    write_control_config()
    write_nodemgr_config()

@hooks.hook("contrail-ifmap-relation-changed")
def contrail_ifmap_changed():
    creds = relation_get("creds")
    creds = json.loads(creds) if creds else {}
    if local_unit() not in creds:
        log("Relation not ready")
        return
    contrail_ifmap_relation()
    config["contrail-ifmap-ready"] = True
    add_control()

@hooks.hook("contrail-ifmap-relation-departed")
@hooks.hook("contrail-ifmap-relation-broken")
def contrail_ifmap_departed():
    if not units("contrail-ifmap"):
        remove_control()
        config["contrail-ifmap-ready"] = False
    contrail_ifmap_relation()

@restart_on_change({"/etc/contrail/contrail-control.conf": ["contrail-control"],
                    "/etc/contrail/control-node.conf": ["contrail-control"]})
def contrail_ifmap_relation():
    write_control_config()

@hooks.hook("control-node-relation-joined")
def control_node_joined():
    settings = { "private-address": control_network_ip() }
    relation_set(relation_settings=settings)

@hooks.hook("identity-admin-relation-changed")
def identity_admin_changed():
    if not relation_get("service_hostname"):
        log("Relation not ready")
        return
    write_vnc_api_config()
    config["identity-admin-ready"] = True
    add_control()

@hooks.hook("identity-admin-relation-departed")
@hooks.hook("identity-admin-relation-broken")
def identity_admin_departed():
    if not units("identity-admin"):
        remove_control()
        config["identity-admin-ready"] = False
    write_vnc_api_config()

@hooks.hook()
def install():
    configure_sources(True, "install-sources", "install-keys")
    apt_upgrade(fatal=True, dist=True)
    apt_install(PACKAGES, fatal=True)
    utils.CONTRAIL_VERSION = dpkg_version("contrail-control")
    fix_permissions()
    fix_nodemgr()

def main():
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log("Unknown hook {} - skipping.".format(e))

def remove_control():
    if config_get("control-provisioned"):
        unprovision_control()
        config["control-provisioned"] = False

@hooks.hook("upgrade-charm")
def upgrade_charm():
    write_control_config()
    write_nodemgr_config()
    service_restart("supervisor-control")

@restart_on_change({"/etc/contrail/contrail-control.conf": ["contrail-control"]})
def write_config():
    write_control_config()

if __name__ == "__main__":
    main()
